%% /u/beebe/src/lex/cweb/expr3.ch, Mon Jun  3 07:08:11 1991
%% Edit by Nelson H. F. Beebe <beebe@rios.math.utah.edu>

This is the change file to convert expr into
expr3, a version of the expression parser that
recognizes power and factorial (gamma function)
operators.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
We need one more macro than before.  We also want
to see in the TeX output only the changes from
the original program

@x
\def\ctangle{\.{ctangle}}
@y
\let\maybe=\iffalse

\def\caret{\char`\^}

\def\ctangle{\.{ctangle}}
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Update the running page headers.

@x
\def\title{EXPR}  % used in the page headers
@y
\def\title{EXPR3} % used in the page headers
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Update the document title page.
@x
  \centerline{\titlefont
    {\ttitlefont EXPR}---A
    Simple Expression Parser}
@y
  \centerline{\titlefont
    {\ttitlefont EXPR}---A
    Simple Expression Parser}
  \centerline{Version 3}
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Here are the initial grammar modifications.

@x
\+        T & : &T * F\cr
\+           & \vbar & T / F\cr
\+           & \vbar & F\cr
@y
\+        T & : &T * P\cr
\+           & \vbar & T / P\cr
\+           & \vbar & P\cr

\+        P & : &G \caret P\cr
\+           & \vbar & G\cr

\+        G & : &F !\cr
\+           & \vbar & F\cr
@z

Here are the final (non-left-recursive) grammar
changes.

@x
\+        T & : &F T'\cr

\+        T' & : &* F\cr
\+            & \vbar & / F\cr
\+            & \vbar & \langle empty\rangle\cr
@y
\+        T & : &P T'\cr

\+        T' & : &* P\cr
\+            & \vbar & / P\cr
\+            & \vbar & \langle empty\rangle\cr

\+        P & : &G \caret P\cr
\+           & \vbar & G\cr

\+        G & : &F !\cr
\+           & \vbar & F\cr
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
We need two more function declarations.

@x
double  factor @, ARGS((void));@/
@y
double  factor @, ARGS((void));@/
double  factorial @, ARGS((void));@/
@z

@x
int     peeknb @, ARGS((void));@/
@y
int     peeknb @, ARGS((void));@/
double  power @, ARGS((void));@/
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Add a no-op change to work around a TeX error.
@x
@* The expression parser.
@y
@* The expression parser.
@z



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Add a section describing the implementation of
the power rule.

@x
@ The factor rule, F, in the grammar
@y
@ The power rule, P, in the grammar

\+        P & : &G \caret P\cr
\+           & \vbar & G\cr

\noindent 
translates to a |for(;;)| loop to handle operator
lookahead.  Return occurs when no further power
operators are found.

@c @<Parse a power@>

@ @<Parse a power@>=
double
power()
{
    double p = factorial();

    for (;;)
    {
        if (peeknb() == '^')
            nextnb(), p = pow(p,power());
        else
            return (p);
    }
}

@ The factor rule, F, in the grammar
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The last part of the grammar needs a new section.

@x
@ The third part of the expression grammar defines
@y
@ The last part of the grammar,

\+        G & : &F !\cr
\+           & \vbar & F\cr

requires only a single-token lookahead to implement.

@c @<Parse a gamma or factorial@>

@ @<Parse a gamma or factorial@>=
double
factorial()
{
    double g = factor();

    if (peeknb() == '!')
	nextnb(), g = gamma(g + 1.0), g = signgam * exp(g);
    return (g);
}

@ The third part of the expression grammar defines
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
In term(), we now get powers instead of factors.

@x
    double t = factor();
@y
    double t = power();
@z

@x
        if (peeknb() == '*')
            nextnb(), t *= factor();
        else if (peeknb() == '/')
            nextnb(), t /= factor();
@y
        if (peeknb() == '*')
            nextnb(), t *= power();
        else if (peeknb() == '/')
            nextnb(), t /= power();
@z

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Finally, we have a different file name.

@x
\input expr.toc % Make the title page and table of contents
@y
\input expr3.toc % Make the title page and table of contents
@z
